"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SessionCompiler = void 0;
const fs = __importStar(require("fs/promises"));
const jszip_1 = __importDefault(require("jszip"));
const crypto_1 = require("crypto");
const reverseShortEventTypes = {
    S: 'scroll',
    C: 'click',
    // TODO to be completed
};
/** Class implementing the session compilation process. */
class SessionCompiler {
    constructor(context, chunks, sessionId, userId) {
        this.chunks = [];
        this.loms = {};
        this.timeline = [];
        this.previousFaultyTs = 0;
        this.lomDeduplicationMapping = {};
        this.lomCounter = 1;
        this.context = context;
        this.chunks = chunks;
        this.session = { sessionId, timeStamp: this.chunks[0].ts, userId, loms: this.loms, timeline: this.timeline };
        this.currentChunk = 0;
    }
    static create(context) {
        return __awaiter(this, void 0, void 0, function* () {
            let sessionId;
            let userId;
            const chunks = [];
            yield Promise.all(context.sourceFiles.map((file) => __awaiter(this, void 0, void 0, function* () {
                const chunk = JSON.parse(yield fs.readFile(file, 'utf8'));
                if (!sessionId)
                    sessionId = chunk.sid;
                else if (sessionId !== chunk.sid)
                    throw new Error(`Inconsistent session IDs in ${file}, ${sessionId} expected, but was ${chunk.sid}`);
                if (!userId)
                    userId = chunk.uid;
                chunks.push(chunk);
            })));
            if (!sessionId)
                throw new Error(`Empty session`);
            chunks.sort((a, b) => a.ts - b.ts);
            return new SessionCompiler(context, chunks, sessionId, userId);
        });
    }
    relativizeTimestampToSession(ts) {
        if (ts >= 0)
            return Math.round(ts + this.chunks[this.currentChunk].ts - this.session.timeStamp);
        // Fix negative timestamps caused by faulty event capture
        let docTs = ts + this.chunks[this.currentChunk].ts;
        if (docTs < this.previousFaultyTs) {
            docTs += this.previousFaultyTs;
        }
        this.previousFaultyTs = docTs;
        return Math.round(docTs);
    }
    addToTimeline(element) {
        if (this.timeline.length === 0) {
            this.timeline.push(element);
            return;
        }
        let index = this.timeline.length;
        while (index > 0 && this.timeline[index - 1].timeStamp > element.timeStamp)
            index -= 1;
        this.timeline.splice(index, 0, element);
    }
    createExplorationEvent(ee) {
        const [ts, t, scrollPos, mousePosOrZoneId = undefined, zoneId = undefined] = ee.split(':');
        const [sx, sy] = scrollPos.split(',');
        const scrollPosition = { x: Number(sx), y: Number(sy) };
        const timeStamp = this.relativizeTimestampToSession(Number(ts));
        const type = reverseShortEventTypes[t] || t;
        let mousePosition;
        let focusedZoneId = zoneId;
        if (mousePosOrZoneId) {
            if (mousePosOrZoneId.includes(',')) {
                const [mx, my] = mousePosOrZoneId.split(',');
                mousePosition = { x: Number(mx), y: Number(my) };
            }
            else
                focusedZoneId = mousePosOrZoneId;
        }
        return { t: 'Exploration', lom: '', type, timeStamp, scrollPosition, mousePosition, focusedZoneId };
    }
    createActionEvent(ae) {
        const [ts, t, zoneId, mousePos = undefined] = ae.split(':');
        const timeStamp = this.relativizeTimestampToSession(Number(ts));
        const type = reverseShortEventTypes[t] || t;
        let mousePosition;
        if (mousePos) {
            const [mx, my] = mousePos.split(',');
            mousePosition = { x: Number(mx), y: Number(my) };
        }
        return { t: 'Action', lom: '', modifiers: 0, type, timeStamp, zoneId, mousePosition };
    }
    findSimilarLom(lom) {
        function isSimilarTo(otherLom, tolerance = 20) {
            return (Math.abs(lom.pageWidth - otherLom.pageWidth) <= tolerance &&
                Math.abs(lom.pageHeight - otherLom.pageHeight) <= tolerance &&
                areSimilar(lom.root, otherLom.root, tolerance));
        }
        function areSimilar(z1, z2, tolerance) {
            return (areCoincident(z1.bounds, z2.bounds, tolerance) &&
                z1.children.length === z2.children.length &&
                z1.children.every((c, i) => areSimilar(c, z2.children[i], tolerance)));
        }
        function areCoincident(b1, b2, tolerance) {
            return (Math.abs(b1.x - b2.x) <= tolerance &&
                Math.abs(b1.y - b2.x) <= tolerance &&
                Math.abs(b1.width - b2.width) <= tolerance &&
                Math.abs(b1.height - b2.height) <= tolerance);
        }
        // eslint-disable-next-line no-restricted-syntax
        for (const lomId of Object.keys(this.loms)) {
            if (isSimilarTo(this.loms[lomId]))
                return lomId;
        }
        return undefined;
    }
    compile() {
        return __awaiter(this, void 0, void 0, function* () {
            function createLomTransition(id, ts) {
                return { t: 'Transition', target: id, timeStamp: ts };
            }
            function createLom(lom) {
                function createZone(z) {
                    var _a;
                    const result = {
                        zoneId: z.id || (0, crypto_1.randomUUID)(),
                        transitions: [],
                        bounds: { x: z.b[0], y: z.b[1], width: z.b[2], height: z.b[3] },
                        style: z.s ? { background: z.s.bg, border: z.s.b } : undefined,
                        children: [],
                    };
                    (_a = z.c) === null || _a === void 0 ? void 0 : _a.forEach((c) => result.children.push(createZone(c)));
                    return result;
                }
                return { pageWidth: lom.w, pageHeight: lom.h, root: createZone(lom.r), title: lom.t, url: lom.u };
            }
            function findZone(lom, id) {
                function doFindZone(zone) {
                    if (zone.zoneId === id)
                        return zone;
                    // eslint-disable-next-line no-restricted-syntax
                    for (const c of zone.children) {
                        const found = doFindZone(c);
                        if (found)
                            return found;
                    }
                    return undefined;
                }
                return doFindZone(lom.root);
            }
            // Step 1 : Build the session timeline
            this.chunks.forEach((chunk, index) => {
                var _a, _b, _c;
                this.currentChunk = index;
                (_a = chunk.loms) === null || _a === void 0 ? void 0 : _a.forEach((lomOrRef) => {
                    const lomTs = this.relativizeTimestampToSession(lomOrRef.ts);
                    if ('r' in lomOrRef) {
                        const lom = createLom(lomOrRef);
                        const similarLomId = this.findSimilarLom(lom);
                        if (similarLomId) {
                            if (lomOrRef.id)
                                this.lomDeduplicationMapping[lomOrRef.id] = similarLomId;
                            this.addToTimeline(createLomTransition(similarLomId, lomTs));
                        }
                        else {
                            // eslint-disable-next-line no-plusplus
                            const lomId = `l${String(this.lomCounter++).padStart(3, '0')}`;
                            if (lomOrRef.id)
                                this.lomDeduplicationMapping[lomOrRef.id] = lomId;
                            this.loms[lomId] = lom;
                            this.addToTimeline(createLomTransition(lomId, lomTs));
                        }
                    }
                    else
                        this.addToTimeline(createLomTransition(this.lomDeduplicationMapping[lomOrRef.ref], lomTs));
                });
                (_b = chunk.ee) === null || _b === void 0 ? void 0 : _b.forEach((ee) => this.addToTimeline(this.createExplorationEvent(ee)));
                (_c = chunk.ae) === null || _c === void 0 ? void 0 : _c.forEach((ae) => this.addToTimeline(this.createActionEvent(ae)));
            });
            // Step 2 : Detect LOM transitions
            let curLomId;
            let lastActionEvent;
            let lastExplorationEvent;
            this.timeline.forEach((event) => {
                var _a;
                switch (event.t) {
                    case 'Transition':
                        if (curLomId) {
                            if (lastActionEvent && lastActionEvent.timeStamp > event.timeStamp - 1000) {
                                const fromZone = findZone(this.loms[curLomId], lastActionEvent.zoneId);
                                (_a = fromZone === null || fromZone === void 0 ? void 0 : fromZone.transitions) === null || _a === void 0 ? void 0 : _a.push({ targetLom: event.target, eventType: 'click' });
                            }
                            else if (lastExplorationEvent) {
                                // TODO retrieve zone for hover events triggering a transition
                                // const fromZone = findZone(this.loms[curLomId], lastExplorationEvent.zoneId);
                                // fromZone?.transitions?.push({ targetLom: event.target, eventType: 'hover' });
                            }
                        }
                        curLomId = event.target;
                        break;
                    case 'Action':
                        // eslint-disable-next-line no-param-reassign
                        event.lom = curLomId || '';
                        lastActionEvent = event;
                        break;
                    case 'Exploration':
                        // eslint-disable-next-line no-param-reassign
                        event.lom = curLomId || '';
                        lastExplorationEvent = event;
                        break;
                    default:
                }
            });
            // Step 3 : Write resulting session object to output directory
            const sessionJSON = JSON.stringify(this.session);
            yield fs.mkdir(this.context.outputDirectory, { recursive: true });
            yield fs.writeFile(`${this.context.outputDirectory}/session.json`, sessionJSON, 'utf8');
            // Step 4 : Build replay HTML site
            const generatedHtmlFiles = {};
            if (this.context.generateReplicaSite) {
                // TODO Build replica site
                yield fs.mkdir(`${this.context.outputDirectory}/replay`, { recursive: true });
                const indexFile = `${this.context.outputDirectory}/replay/index.html`;
                yield fs.writeFile(indexFile, 'TEST', 'utf8');
                generatedHtmlFiles['replay/index.html'] = indexFile;
            }
            // Step 5 : Create archive
            if (this.context.generateArchive) {
                const zip = new jszip_1.default();
                zip.file('session.json', sessionJSON);
                Object.entries(generatedHtmlFiles).forEach(([path, file]) => {
                    zip.file(path, fs.readFile(file, 'utf8'));
                });
                const buffer = yield zip.generateAsync({ type: 'nodebuffer', streamFiles: true });
                yield fs.writeFile(`${this.context.outputDirectory}/session.zip`, buffer);
            }
        });
    }
}
exports.SessionCompiler = SessionCompiler;
